from functools import reduce
from .TrainingUtils import *
from ..Exceptions import *

class MultiInstrumentTickTrainingUtils(TickTrainingUtils):
	def __init__(self, userConfig, modelConfig):
		TickTrainingUtils.__init__(self, userConfig)
		self.factorsInfo = dict()
		for instConfig in modelConfig['Instruments']:
			conf = dict()
			#因子库因子
			if 'LibFactors' in instConfig:
				conf['FactorLibInfo'] = {lib['LibName']:{'Fields':[f['Name'] for f in lib['Factors']]} for lib in instConfig['LibFactors']}
			#本地因子集
			if 'LocalFactors' in instConfig:
				conf['LocalSetInfo'] = {lSet['SetName']:{'RootDir':lSet['RootDir'],'Fields':[f['Name'] for f in lSet['Factors']]} for lSet in instConfig['LocalFactors']}
			self.factorsInfo[instConfig['Name']] = conf

	def CompleteFactorsFrame(self, tradingDate):
		try:
			instRawFrames = [self.InstrumentFactorsFrame(instrument, tradingDate) for instrument in self.factorsInfo.keys()]
			if len(instRawFrames) == 0:
				raise NoDataError('error, no instrument config for model, please check model config file')
			elif len(instRawFrames) == 1:
				consFrame = instRawFrames[0]
			else:
				consFrame = reduce(lambda left,right: pd.merge(left, right, on=TickTrainingUtils.alignField), instRawFrames)
			return consFrame
		except Exception as e:
			print(e)
			raise NoDataError('error, failed to get consolidated raw factors frame on %s' % tradingDate)

	def InstrumentFactorsFrame(self, instrument, tradingDate):
		try:
			frames = list()
			dFields = list()
			if 'FactorLibInfo' in self.factorsInfo[instrument]:
				for libName in self.factorsInfo[instrument]['FactorLibInfo'].keys():
					dFields.extend(self.factorsInfo[instrument]['FactorLibInfo'][libName]['Fields'])
					frames.append(self.GetLibFactorsData(libName, instrument, tradingDate, self.factorsInfo[instrument]['FactorLibInfo'][libName]['Fields']))
			if 'LocalSetInfo' in self.factorsInfo[instrument]:
				for setName in self.factorsInfo[instrument]['LocalSetInfo'].keys():
					dFields.extend(self.factorsInfo[instrument]['LocalSetInfo'][setName]['Fields'])
					frames.append(self.GetLocalFactorsData(setName, self.factorsInfo[instrument]['LocalSetInfo'][setName]['RootDir'], instrument, tradingDate, self.factorsInfo[instrument]['LocalSetInfo'][setName]['Fields']))
			renamingDict = {f:'%s_%s' % (instrument, f) for f in dFields}
			if len(frames) == 0:
				raise NoDataError('error, no data field specified for %s, please check model config file' % instrument)
			elif len(frames) == 1:
				consFrame = frames[0]
			else:
				consFrame = reduce(lambda left,right: pd.merge(left, right, on=TickTrainingUtils.alignField), frames)
			consFrame.rename(columns=renamingDict, inplace=True)
			return consFrame
		except Exception as e:
			print(e)
			raise NoDataError('error, failed to get raw factors frame for %s on %s' % (instrument, tradingDate))



# -*- coding: utf-8 -*-
from dateutil.parser import parse
from .DatabaseManager import *
from .Exceptions import *

def StockIdToTableName(stockId):
	parts = stockId.lower().split(".")
	return parts[1] +  parts[0]

def IsValidDate(strdate):
    try:
        if "-" in strdate:
            datetime.datetime.strptime(strdate, "%Y-%m-%d")
        else:
            datetime.datetime.strptime(strdate, "%Y%m%d")
        return True
    except:
        return False

#将列表拆分成长度尽可能平均的子列表
def Chunks(arr, m):
	n = int(np.ceil(len(arr) / float(m)))
	return [arr[i:i + n] for i in range(0, len(arr), n)]

def MinuteIdToTimeOfDay(minuteId):
	return time.strftime('%H:%M:%S', time.gmtime(60*minuteId))

def TimestampToMinuteId(timestamp):
	return 60*timestamp.hour+timestamp.minute

def TimestampToSecondsCount(timestamp):
	return 3600*timestamp.hour + 60*timestamp.minute + timestamp.second

def TimeOfDayToMinuteId(todStr):
	todParts = todStr.split(":")
	return 60*int(todParts[0]) + int(todParts[1])

def TimeOfDayToSecondsCount(todStr):
	todParts = todStr.split(":")
	return 3600*int(todParts[0]) + 60*int(todParts[1]) + int(todParts[2])

def TickOfSecond(millisec):
	if millisec < 500:
		return 0
	else:
		return 1

#每秒2次的Tick数据时间戳转化为对齐字段TickLabel
def T2TimeStampToTickLabel(todStr, millisec):
	return 2*TimeOfDayToSecondsCount(todStr) + TickOfSecond(millisec)

def TimeStampToT2TickLabel(todStr, millisec):
	return 2*TimeOfDayToSecondsCount(todStr) + TickOfSecond(millisec)

#N秒一次的Tick数据时间戳转化为对齐字段TickLabel
def TimeStampToNSecTickLabel(timestamp, n):
	return n*int((TimestampToSecondsCount(timestamp)+n-1)/n)

#N秒一次的Tick数据时间字符串转化为对齐字段TickLabel
def TimeOfDayToNSecTickLabel(todStr, n):
	return n*int((TimeOfDayToSecondsCount(todStr)+n-1)/n)

def VolumeWeighedPrice(bidPrice1, bidVolume1, askPrice1, askVolume1):
	return (bidPrice1*askVolume1 + askPrice1*bidVolume1)/(bidVolume1 + askVolume1)

class MarketDataUtils(object):
	dbManager = MarketDatabaseManager()

#期货数据接口
class FuturesDataUtils(MarketDataUtils):
	def __init__(self, dataSource):
		if not dataSource in ['cffex-l1', 'shfe-l1', 'ine-l1', 'dce-l1', 'czce-l1', 'cffex-l2', 'shfe-l2', 'ine-l2', 'dce-l2', 'czce-l2']:
			print('HiggsBoom FuturesDataUtils: error, unknown data source %s' % dataSource)
			sys.exit(-1)
		self.dataSource = dataSource
		self.exchangeId = dataSource.split('-')[0].upper()
		print('HiggsBoom: initializing futures data engine for %s' % self.dataSource)

	def TickDataFrame(self, instrument, tradingDate):
		try:
			tradingDate = str(parse(tradingDate).date()).replace('-', '')
			instrumentId = instrument
			if '_V' in instrument:
				parts = instrument.split('_V')
				summaryDf = self.DailySummaryFrame(parts[0], tradingDate).sort_values(by=['TotalVolume'], ascending=False)
				if summaryDf.shape[0] == 0:
					raise NoDataError('no instrument data for %s on %s' % (parts[0], tradingDate))
				instrumentId = list(summaryDf['InstrumentId'])[int(parts[1])]
			tableName = '%s_%s' % (self.exchangeId, tradingDate)
			selectSql = "SELECT * FROM %s WHERE instrumentId='%s'" % (tableName, instrumentId)
			tickFrame = pd.DataFrame(MarketDataUtils.dbManager.GetQueryResult(self.dataSource, selectSql), columns=[x[0].upper() + x[1:] for x in MarketDataUtils.dbManager.GetTableColumns(self.dataSource, tableName)])
			return tickFrame
		except:
			raise NoDataError("error, failed to get tick data for %s on %s" % (instrument, tradingDate))

	def MinuteDataFrame(self, instrument, tradingDate):
		try:
			tickFrame = self.TickDataFrame(instrument, tradingDate)
			tickFrame['MinuteID'] = tickFrame['UpdateTime'].apply(TimeOfDayToMinuteId)
			minGroup = tickFrame.groupby('MinuteID')
			minFrame = minGroup.agg({'LastPrice':['first', 'max', 'min', 'last'], 'Volume':['last'], 'Turnover':['last']})
			minFrame.reset_index(inplace=True)
			minFrame.columns = [''.join(col).strip() for col in minFrame.columns.values]
			minFrame.rename(columns={'LastPricefirst':'Open', 'LastPricemax':'High', 'LastPricemin':'Low','LastPricelast':'Close', 'Volumelast':'AccVolume','Turnoverlast':'AccAmt'}, inplace=True)
			minFrame['Volume'] = minFrame['AccVolume'] - minFrame['AccVolume'].shift(1)
			minFrame['Amt'] = minFrame['AccAmt'] - minFrame['AccAmt'].shift(1)
			minFrame['Timestamp'] = minFrame['MinuteID'].apply(MinuteIdToTimeOfDay)
			return minFrame[['Timestamp', 'MinuteID', 'Open', 'High', 'Low', 'Close', 'Volume', 'Amt']]
		except NoDataError as e:
			raise NoDataError("error, failed to get minute data for %s on %s" % (instrument, tradingDate))

	def DailySummaryFrame(self, productId, tradingDate):
		tradingDate = str(parse(tradingDate).date()).replace('-', '')
		selectSql = "SELECT * FROM %s_summary WHERE ProductId='%s' AND ProductType='FUT' AND TradingDate='%s'" % (self.exchangeId, productId, tradingDate)
		return pd.DataFrame(MarketDataUtils.dbManager.GetQueryResult(self.dataSource, selectSql), columns=MarketDataUtils.dbManager.GetTableColumns(self.dataSource, '%s_summary' % self.exchangeId))

	def TradingDays(self, productId):
		selectSql = "SELECT DISTINCT TradingDate FROM %s_summary WHERE ProductId='%s' AND ProductType='FUT'" % (self.exchangeId, productId)
		tdList = [x[0] for x in MarketDataUtils.dbManager.GetQueryResult(self.dataSource, selectSql)]
		tdList.sort()
		return tdList

#A股Tick级数据接口
class AStockTickDataUtils(MarketDataUtils):
	def __init__(self):
		self.dataSource = 'astock-tick'
		print('HiggsBoom: initializing tick data engine for A-stocks')

	def InstrumentList(self, instType, tradingDate):
		selectSql = "SELECT InstrumentId FROM tick_summary WHERE InstrumentType='%s' AND TradingDate='%s'" % (instType, str(parse(tradingDate).date()))
		return [x[0] for x in MarketDataUtils.dbManager.GetQueryResult(self.dataSource, selectSql)]

	def TradingDays(self):
		selectSql = "SELECT DISTINCT TradingDate FROM tick_summary"
		tdList = [str(x[0]) for x in MarketDataUtils.dbManager.GetQueryResult(self.dataSource, selectSql)]
		tdList.sort()
		return tdList

	def TickDataFrame(self, instrument, tradingDate):
		tableName = StockIdToTableName(instrument)
		selectSql = "SELECT * FROM %s WHERE DATE(Timestamp)='%s'" % (tableName, str(parse(tradingDate).date()))
		tickFrame = pd.DataFrame(MarketDataUtils.dbManager.GetQueryResult(self.dataSource, selectSql), columns=[x[0].upper() + x[1:] for x in MarketDataUtils.dbManager.GetTableColumns(self.dataSource, tableName)])
		if tickFrame.shape[0] == 0:
			raise NoDataError("error, failed to get tick data for %s on %s" % (instrument, tradingDate))
		return tickFrame

	def MinuteDataFrame(self, instrument, tradingDate):
		try:
			tickFrame = self.TickDataFrame(instrument, tradingDate)
			tickFrame['MinuteID'] = tickFrame['Timestamp'].apply(TimestampToMinuteId)
			minGroup = tickFrame.groupby('MinuteID')
			minFrame = minGroup.agg({'LastPrice':['first', 'max', 'min', 'last'], 'Volume':['last'], 'Amount':['last']})
			minFrame.reset_index(inplace=True)
			minFrame.columns = [''.join(col).strip() for col in minFrame.columns.values]
			minFrame.rename(columns={'LastPricefirst':'Open', 'LastPricemax':'High', 'LastPricemin':'Low','LastPricelast':'Close', 'Volumelast':'AccVolume','Amountlast':'AccAmt'}, inplace=True)
			minFrame['Volume'] = minFrame['AccVolume'] - minFrame['AccVolume'].shift(1)
			minFrame['Amt'] = minFrame['AccAmt'] - minFrame['AccAmt'].shift(1)
			minFrame['Timestamp'] = minFrame['MinuteID'].apply(MinuteIdToTimeOfDay)
			return minFrame[['Timestamp', 'MinuteID', 'Open', 'High', 'Low', 'Close', 'Volume', 'Amt']]
		except NoDataError as e:
			raise NoDataError("error, failed to get minute data for %s on %s" % (instrument, tradingDate))



from .DatabaseManager import FactorsDatabaseManager

class TrainingUtils(object):
	def __init__(self, config):
		self.factorsLib = config['FactorsLibrary']
		self.factorsDB = FactorsDatabaseManager(config['FactorsLibrary'], config['User'])



from abc import ABC, abstractmethod
from .MarketDataUtils import *
from .Exceptions import *

pd.set_option('mode.chained_assignment', None)

class TickFactorsGeneratorBase(ABC):
	def __init__(self):
		self.tickLabelFrame = self.TickLabelFrame()

	@abstractmethod
	def TickLabelFrame(self):
		return pd.DataFrame()

	def FactorsFrameGenerationProcess(self, instrument, tradingDate):
		#try:
		rawTickFrame = self.RawTickFrame(instrument, tradingDate)
		tickFrame = self.TickFramePreprocessing(rawTickFrame)
		factorsFrame = self.GenerateFactorsFrame(tickFrame)
		return factorsFrame
		#except NoDataError as e:
		#print(e)
		#raise ProcessFailure("error, failed to generate factors frame for %s on %s" % (instrument, tradingDate))

	@abstractmethod
	def RawTickFrame(self, instrument, tradingDate):
		return pd.DataFrame()

	@abstractmethod
	def TickFramePreprocessing(self, rawTickFrame):
		return rawTickFrame

	@abstractmethod
	def GenerateFactorsFrame(self, dataFrame):
		return dataFrame


import os
import numpy as np
import pandas as pd
from dateutil.parser import parse
from .DatabaseManager import FactorsDatabaseManager
from .Exceptions import *

class TrainingUtils(object):
	def __init__(self, userInfo):
		self.dbManager = FactorsDatabaseManager(userInfo)

class TickTrainingUtils(object):
	alignField = 'TickLabel'

	def __init__(self, userInfo):
		TrainingUtils.__init__(self, userInfo)

	def GetLibFactorsData(self, libName, instrument, tradingDate, dataFields):
		try:
			tradingDate = str(parse(tradingDate).date()).replace('-', '')
			tableName = [tb for tb in self.dbManager.FactorTables(libName) if tb.endswith('Factors_%s_%s' % (instrument, tradingDate))]
			assert len(tableName) == 1, 'error, no factors data table for %s on %s in %s' % (instrument, tradingDate, libName)
			availFields = self.dbManager.GetTableColumns(libName, tableName[0])
			sql = "SELECT %s" % self.alignField
			cols = [TickTrainingUtils.alignField]
			for f in dataFields:
				sql += ', %s' % f
				cols.append(f)
				if not f in availFields:
					raise NoDataError('error, field %s not in factors table for %s on %s' % (f, instrumentId, tradingDate))
			sql += ' FROM %s' % tableName[0]
			return pd.DataFrame(self.dbManager.GetQueryResult(libName, sql), columns = cols)
		except Exception as e:
			print(e)
			raise NoDataError('error, failed to get factors frame from library %s for %s on %s' % (libName ,instrument, tradingDate))

	def GetLocalFactorsData(self, setName, rootDir, instrument, tradingDate, dataFields):
		try:
			tradingDate = str(parse(tradingDate).date()).replace('-', '')
			fileName = '%s/%s/%s.csv' % (rootDir, tradingDate, instrument)
			if not os.path.exists(fileName):
				raise NoDataError('error, factor set %s does not contain data file for %s on %s' % (setName, instrument, tradingDate))
			df = pd.read_csv('%s/%s/%s.csv' % (rootDir, tradingDate, instrument))
			for f in dataFields:
				if not f in df.columns:
					raise NoDataError('error, field %s not in factors table for %s on %s' % (f, instrument, tradingDate))
			return df[[TickTrainingUtils.alignField]+dataFields]
		except NoDataError as e:
			print(e)
			raise NoDataError('error, failed to get factors frame from local set for %s on %s' % (setName, instrument, tradingDate))


import sqlalchemy
import sqlalchemy.orm
import pandas as pd

def EngineConfigToCreateStr(config):
	createStr = config['dbEngine']
	if 'dbDriver' in config:
		createStr += '+%s' % config['dbDriver']
	createStr += '://%s' % config['user']
	if 'passwd' in config:
		createStr += ':%s' % config['passwd']
	createStr += '@%s' % config['host']
	if 'port' in config:
		createStr += ':%d' % config['port']
	if 'dbName' in config:
		createStr += '/%s' % config['dbName']
	return createStr

def DataSourceToDatabaseName(dataSource):
	if dataSource in ['cffex-l1', 'shfe-l1', 'ine-l1', 'dce-l1', 'czce-l1']:
		return 'Level1Md'
	elif dataSource in ['cffex-l2', 'shfe-l2', 'ine-l2', 'dce-l2', 'czce-l2']:
		return 'Level2Md'
	elif dataSource == 'astock-daily':
		return 'ASTOCK-DAILY'
	elif dataSource == 'astock-minute':
		return 'ASTOCK-MINUTE'
	elif dataSource == 'astock-tick':
		return 'ASTOCK-TICK'
	else:
		return 'UknownDB'


class MarketDatabaseManager(object):
	dbList = ['Level1Md', 'Level2Md', 'ASTOCK-DAILY', 'ASTOCK-MINUTE', 'ASTOCK-TICK']

	def __init__(self):
		self.dbEngines = dict()
		self.dbSessions = dict()

	def InitDataEngine(self, dbName):
		dbConfig = dict()
		dbConfig['dbEngine'] = 'mysql'
		dbConfig['passwd'] = 'Higgs1401'
		dbConfig['user'] = 'mdreader'
		dbConfig['host'] = '10.214.0.21'
		dbConfig['dbName'] = dbName
		self.dbEngines[dbName] = sqlalchemy.create_engine(EngineConfigToCreateStr(dbConfig), echo=False)
		self.dbSessions[dbName] = sqlalchemy.orm.sessionmaker(bind=self.dbEngines[dbName])()

	def GetTableColumns(self, dataSource, tableName):
		dbName = DataSourceToDatabaseName(dataSource)
		if not dbName in MarketDatabaseManager.dbList:
			print('error, no corresponding database for %s' % dataSource)
			return None
		if not dbName in self.dbEngines.keys():
			self.InitDataEngine(dbName)
		return [x['name'] for x in sqlalchemy.inspect(self.dbEngines[dbName]).get_columns(tableName)]

	def GetQueryResult(self, dataSource, sql):
		dbName = DataSourceToDatabaseName(dataSource)
		if not dbName in MarketDatabaseManager.dbList:
			print('error, no corresponding database for %s' % dataSource)
			return None
		if not dbName in self.dbEngines.keys():
			self.InitDataEngine(dbName)
		return self.dbSessions[dbName].execute(sql).fetchall()

class FactorsDatabaseManager(object):
	def __init__(self, userInfo):
		self.userInfo = userInfo
		self.dbEngines = dict()
		self.dbSessions = dict()
		self.factorTables = dict()

	def InitFactorsLibEngine(self, factorsLib):
		try:
			dbConfig = dict()
			dbConfig['dbEngine'] = 'mysql'
			dbConfig['user'] = self.userInfo['UserName']
			dbConfig['passwd'] = self.userInfo['Password']
			dbConfig['host'] = '10.214.0.21'
			dbConfig['dbName'] = factorsLib
			self.dbEngines[factorsLib] = sqlalchemy.create_engine(EngineConfigToCreateStr(dbConfig), echo=False)
			self.dbSessions[factorsLib] = sqlalchemy.orm.sessionmaker(bind=self.dbEngines[factorsLib])()
			print('HiggsBoom: initialize factors database session for user %s on %s' % (self.userInfo['UserName'], factorsLib))
		except:
			print('HiggsBoom: error, failed to initialize factors database session for user %s on %s' % (self.userInfo['UserName'], factorsLib))

	def FactorTables(self, factorsLib):
		if not factorsLib in self.dbEngines.keys():
			self.InitFactorsLibEngine(factorsLib)
		return [x[0] for x in self.GetQueryResult(factorsLib, 'SHOW TABLES')]

	def GetTableColumns(self, factorsLib, tableName):
		if not factorsLib in self.dbEngines.keys():
			self.InitFactorsLibEngine(factorsLib)
		return [x[0] for x in self.GetQueryResult(factorsLib, 'DESCRIBE %s' % tableName)]

	def GetQueryResult(self, factorsLib, sql):
		if not factorsLib in self.dbEngines.keys():
			self.InitFactorsLibEngine(dbName)
		return self.dbSessions[factorsLib].execute(sql).fetchall()

